from .base import (
    Experiment,
    OnRoleExperiment,
)
from .base_complex import CycledExperiment, GroupExperiment

__all__ = ["CycledExperiment", "Experiment", "GroupExperiment", "OnRoleExperiment"]



from __future__ import annotations

from typing import Literal

from .analyzers.ab import ABAnalyzer
from .comparators import Chi2Test, GroupDifference, GroupSizes, TTest, UTest
from .dataset import TargetRole, TreatmentRole
from .experiments.base import Experiment, OnRoleExperiment
from .ui.ab import ABOutput
from .ui.base import ExperimentShell
from .utils import ABNTestMethodsEnum


class ABTest(ExperimentShell):
    """A class for conducting A/B tests with configurable statistical tests and multiple testing correction.

    This class provides functionality to run A/B tests with options for different statistical tests
    (t-test, u-test, chi-square test) and multiple testing correction methods.

    Args:
        additional_tests (Union[str, List[str], None], optional): Statistical test(s) to run in addition to
            the default group difference calculation. Valid options are "t-test", "u-test", and "chi2-test".
            Can be a single test name or list of test names. Defaults to ["t-test"].
        multitest_method (str, optional): Method to use for multiple testing correction. Valid options are:
            "bonferroni", "sidak", "holm-sidak", "holm", "simes-hochberg", "hommel", "fdr_bh", "fdr_by",
            "fdr_tsbh", "fdr_tsbhy", "quantile". Defaults to "holm".
         For more information refer to the statsmodels documentation:
         <https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html>

    Examples:
        Basic A/B test with default t-test:
        >>> ab_test = ABTest()
        >>> results = ab_test.execute(data)

        A/B test with multiple statistical tests:
        >>> ab_test = ABTest(
        ...     additional_tests=["t-test", "chi2-test"],
        ...     multitest_method="bonferroni"
        ... )
        >>> results = ab_test.execute(data)

        A/B test with single non-default test:
        >>> ab_test = ABTest(
        ...     additional_tests="u-test",
        ...     multitest_method="fdr_bh"
        ... )
        >>> results = ab_test.execute(data)
    """

    @staticmethod
    def _make_experiment(additional_tests, multitest_method):
        """Creates an experiment configuration with specified statistical tests.

        Args:
            Args:
        additional_tests (Union[str, List[str], None], optional): Statistical test(s) to run in addition to
            the default group difference calculation. Valid options are "t-test", "u-test", and "chi2-test".
            Can be a single test name or list of test names. Defaults to ["t-test"].
        multitest_method (str, optional): Method to use for multiple testing correction. Valid options are:
            "bonferroni", "sidak", "holm-sidak", "holm", "simes-hochberg", "hommel", "fdr_bh", "fdr_by",
            "fdr_tsbh", "fdr_tsbhy", "quantile". Defaults to "holm".
         For more information refer to the statsmodels documentation:
         <https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html>

        Returns:
            Experiment: Configured experiment object with specified tests and correction method.
        """
        test_mapping = {
            "t-test": TTest(compare_by="groups", grouping_role=TreatmentRole()),
            "u-test": UTest(compare_by="groups", grouping_role=TreatmentRole()),
            "chi2-test": Chi2Test(compare_by="groups", grouping_role=TreatmentRole()),
        }
        on_role_executors = [GroupDifference(grouping_role=TreatmentRole())]
        additional_tests = ["t-test"] if additional_tests is None else additional_tests
        additional_tests = (
            additional_tests
            if isinstance(additional_tests, list)
            else [additional_tests]
        )
        for i in additional_tests:
            on_role_executors += [test_mapping[i]]
        return Experiment(
            executors=[
                GroupSizes(grouping_role=TreatmentRole()),
                OnRoleExperiment(
                    executors=on_role_executors,
                    role=TargetRole(),
                ),
                ABAnalyzer(
                    multitest_method=(
                        ABNTestMethodsEnum(multitest_method)
                        if multitest_method
                        else None
                    )
                ),
            ]
        )

    def __init__(
        self,
        additional_tests: (
            Literal["t-test", "u-test", "chi2-test"]
            | list[Literal["t-test", "u-test", "chi2-test"]]
            | None
        ) = None,
        multitest_method: (
            Literal[
                "bonferroni",
                "sidak",
                "holm-sidak",
                "holm",
                "simes-hochberg",
                "hommel",
                "fdr_bh",
                "fdr_by",
                "fdr_tsbh",
                "fdr_tsbhy",
                "quantile",
            ]
            | None
        ) = "holm",
    ):
        super().__init__(
            experiment=self._make_experiment(additional_tests, multitest_method),
            output=ABOutput(),
        )



from typing import Union

from ..analyzers.ab import ABAnalyzer
from ..comparators import GroupDifference, GroupSizes
from ..dataset import Dataset, ExperimentData, StatisticRole, TreatmentRole
from ..reporters.ab import ABDatasetReporter
from ..utils import ID_SPLIT_SYMBOL, ExperimentDataEnum
from .base import Output


class ABOutput(Output):
    multitest: Union[Dataset, str]
    sizes: Dataset

    def __init__(self):
        self._groups = []
        super().__init__(resume_reporter=ABDatasetReporter())

    def _extract_multitest_result(self, experiment_data: ExperimentData):
        multitest_id = experiment_data.get_one_id(
            ABAnalyzer, ExperimentDataEnum.analysis_tables
        )
        if multitest_id and "MultiTest" in multitest_id:
            self.multitest = experiment_data.analysis_tables[multitest_id]
        else:
            self.multitest = (
                "There was less than three groups or multitest method wasn't provided"
            )

    def _extract_differences(self, experiment_data: ExperimentData):
        targets = []
        groups = []
        ids = experiment_data.get_ids(
            GroupDifference,
            searched_space=ExperimentDataEnum.analysis_tables,
        )["GroupDifference"]["analysis_tables"]
        self._groups = list(
            experiment_data.groups[
                experiment_data.ds.search_columns(TreatmentRole())[0]
            ].keys()
        )[1:]
        for i in self._groups:
            groups += [i] * len(ids)
        diff = Dataset.create_empty()
        for i in range(len(ids)):
            diff = diff.append(experiment_data.analysis_tables[ids[i]])
            targets += [ids[i].split(ID_SPLIT_SYMBOL)[-1]]
        return diff.add_column(groups, role={"group": StatisticRole()}).add_column(
            targets * len(self._groups), role={"feature": StatisticRole()}
        )

    def _extract_sizes(self, experiment_data: ExperimentData):
        ids = experiment_data.get_ids(
            GroupSizes,
            searched_space=ExperimentDataEnum.analysis_tables,
        )["GroupSizes"]["analysis_tables"]
        self.sizes = experiment_data.analysis_tables[ids[0]].add_column(
            self._groups, role={"group": StatisticRole()}
        )

    def extract(self, experiment_data: ExperimentData):
        super().extract(experiment_data)
        self._extract_differences(experiment_data)
        self._extract_multitest_result(experiment_data)
        self._extract_sizes(experiment_data)



from __future__ import annotations

from abc import ABC

from ..utils import (
    CategoricalTypes,
    DefaultRoleTypes,
    RoleNameType,
    TargetRoleTypes,
)


class ABCRole(ABC):

    _role_name: RoleNameType = "Abstract"

    def __init__(self, data_type: DefaultRoleTypes | None = None):
        self.data_type = data_type

    @property
    def role_name(self) -> str:
        return self._role_name

    def __repr__(self) -> str:
        return f"{self._role_name}({self.data_type})"


class InfoRole(ABCRole):

    _role_name: RoleNameType = "Info"


class StratificationRole(ABCRole):

    _role_name: RoleNameType = "Stratification"

    def __init__(self, data_type: CategoricalTypes | None = None):
        super().__init__(data_type)


class GroupingRole(ABCRole):

    _role_name: RoleNameType = "Grouping"

    def __init__(self, data_type: CategoricalTypes | None = None):
        super().__init__(data_type)


class TreatmentRole(ABCRole):

    _role_name: RoleNameType = "Treatment"


class TargetRole(ABCRole):

    _role_name: RoleNameType = "Target"

    def __init__(self, data_type: TargetRoleTypes | None = None):
        super().__init__(data_type)


class FeatureRole(ABCRole):

    _role_name: RoleNameType = "Feature"


class PreTargetRole(ABCRole):

    _role_name: RoleNameType = "PreTarget"

    def __init__(self, data_type: TargetRoleTypes | None = None):
        super().__init__(data_type)


class StatisticRole(ABCRole):

    _role_name: RoleNameType = "Statistic"


class ResumeRole(ABCRole):
    _role_name = "Resume"


class FilterRole(ABCRole):
    _role_name: RoleNameType = "Filter"


class ConstGroupRole(ABCRole):
    _role_name: RoleNameType = "ConstGroup"


# ___________________________________________________________________________________________
class TempRole(ABCRole):
    _role_name: RoleNameType = "Temp"


class TempTreatmentRole(TempRole, TreatmentRole):

    _role_name: RoleNameType = "TempTreatment"


class TempTargetRole(TempRole, TargetRole):

    _role_name: RoleNameType = "TempTarget"


class TempGroupingRole(TempRole, GroupingRole):

    _role_name: RoleNameType = "TempGrouping"


class DefaultRole(ABCRole):

    _role_name: RoleNameType = "Default"


class ReportRole(ABCRole):

    _role_name: RoleNameType = "Report"


# ___________________________________________________________________________________________
class AdditionalRole(ABCRole):

    _role_name: RoleNameType = "Additional"


class AdditionalTreatmentRole(AdditionalRole):

    _role_name: RoleNameType = "AdditionalTreatment"


class AdditionalGroupingRole(AdditionalRole):

    _role_name: RoleNameType = "AdditionalGrouping"


class AdditionalTargetRole(AdditionalRole):

    _role_name: RoleNameType = "AdditionalTarget"


class AdditionalPreTargetRole(AdditionalRole):

    _role_name: RoleNameType = "AdditionalPreTarget"


class AdditionalMatchingRole(AdditionalRole):

    _role_name: RoleNameType = "AdditionalMatching"


default_roles: dict[RoleNameType, ABCRole] = {
    "info": InfoRole(),
    "default": DefaultRole(),
    "feature": FeatureRole(),
    "treatment": TreatmentRole(),
    "grouping": GroupingRole(),
    "target": TargetRole(),
    "pretarget": PreTargetRole(),
    "stratification": StratificationRole(),
    "statistic": StatisticRole(),
    "filter": FilterRole(),
    "constgroup": ConstGroupRole(),
    "additionaltreatment": AdditionalTreatmentRole(),
    "additionalgrouping": AdditionalGroupingRole(),
    "additionaltarget": AdditionalTargetRole(),
    "additionalpretarget": AdditionalPreTargetRole(),
}



from __future__ import annotations

from typing import Literal

from .analyzers.matching import MatchingAnalyzer
from .comparators import KSTest, TTest
from .comparators.distances import MahalanobisDistance
from .dataset import TargetRole, TreatmentRole
from .executor import Executor
from .experiments import GroupExperiment
from .experiments.base import Experiment, OnRoleExperiment
from .ml.faiss import FaissNearestNeighbors
from .operators.operators import Bias, MatchingMetrics
from .reporters.matching import MatchingDatasetReporter
from .ui.base import ExperimentShell
from .ui.matching import MatchingOutput


class Matching(ExperimentShell):
    """A class for performing matching analysis with configurable distance metrics and quality tests.

    This class provides functionality to identify groups that are similar to each other
    using various distance metrics and quality assessment methods.

    Args:
        group_match (bool, optional): Whether to perform group matching. Defaults to False.
        distance (Literal["mahalanobis", "l2"], optional): Distance metric to use for matching.
            Options are "mahalanobis" or "l2". Defaults to "mahalanobis".
        metric (Literal["atc", "att", "ate"], optional): Type of treatment effect to estimate.
            "atc" = average treatment effect on controls
            "att" = average treatment effect on treated
            "ate" = average treatment effect
            Defaults to "ate".
        bias_estimation (bool, optional): Whether to estimate bias. Defaults to True.
        quality_tests (Union[str, List[str]], optional): Quality tests to perform.
            Options are "smd", "psi", "ks-test", "repeats", "t-test", or "auto".
            Can be a single test or list of tests. Defaults to "auto".

    Examples:
        Basic matching with default settings:
        >>> matching = Matching()
        >>> results = matching.execute(data)

        Matching with L2 distance and specific quality tests:
        >>> matching = Matching(
        ...     distance="l2",
        ...     quality_tests=["t-test", "ks-test"]
        ... )
        >>> results = matching.execute(data)

        Group matching with ATT estimation:
        >>> matching = Matching(
        ...     group_match=True,
        ...     metric="att",
        ...     bias_estimation=True
        ... )
        >>> results = matching.execute(data)
    """

    @staticmethod
    def _make_experiment(
            group_match: bool = False,
            distance: Literal["mahalanobis", "l2"] = "mahalanobis",
            metric: Literal["atc", "att", "ate"] = "ate",
            bias_estimation: bool = True,
            quality_tests: (
                    Literal["smd", "psi", "ks-test", "repeats", "t-test", "auto"]
                    | list[Literal["smd", "psi", "ks-test", "repeats", "t-test", "auto"]]
            ) = "auto",
            faiss_mode: Literal["base", "fast", "auto"] = "auto",
    ) -> Experiment:
        """Creates an experiment configuration with specified matching parameters.

        Args:
            group_match (bool, optional): Whether to perform group matching. Defaults to False.
        distance (Literal["mahalanobis", "l2"], optional): Distance metric to use for matching.
            Options are "mahalanobis" or "l2". Defaults to "mahalanobis".
        metric (Literal["atc", "att", "ate"], optional): Type of treatment effect to estimate.
            "atc" = average treatment effect on controls
            "att" = average treatment effect on treated
            "ate" = average treatment effect
            Defaults to "ate".
        bias_estimation (bool, optional): Whether to estimate bias. Defaults to True.
        quality_tests (Union[str, List[str]], optional): Quality tests to perform.
            Options are "ks-test", "t-test", or "auto".
            Can be a single test or list of tests. Defaults to "auto", which performs all tests.

        Returns:
            Experiment: Configured experiment object with specified matching parameters.

        Examples:
            >>> exp = Matching._make_experiment(
            ...     distance="l2",
            ...     metric="att",
            ...     quality_tests=["t-test"]
            ... )
        """
        distance_mapping = {
            "mahalanobis": MahalanobisDistance(grouping_role=TreatmentRole())
        }
        test_mapping = {
            "t-test": TTest(compare_by="groups", grouping_role=TreatmentRole()),
            # "psi": PSI(grouping_role=TreatmentRole(), compare_by="groups"),
            "ks-test": KSTest(grouping_role=TreatmentRole(), compare_by="groups"),
        }
        two_sides = metric == "ate"
        test_pairs = metric == "atc"
        executors: list[Executor] = [
            FaissNearestNeighbors(
                grouping_role=TreatmentRole(),
                two_sides=two_sides,
                test_pairs=test_pairs,
                faiss_mode=faiss_mode,
            )
        ]
        if bias_estimation:
            executors += [
                Bias(grouping_role=TreatmentRole(), target_roles=[TargetRole()]),
            ]
        executors += [
            MatchingMetrics(
                grouping_role=TreatmentRole(),
                target_roles=[TargetRole()],
                metric=metric,
            ),
            MatchingAnalyzer(),
        ]
        if quality_tests != "auto":
            # warnings.warn("Now quality tests aren't supported yet")
            executors += [
                OnRoleExperiment(
                    executors=[test_mapping[test] for test in quality_tests],
                    role=TargetRole(),
                )
            ]
        return (
            Experiment(
                executors=(
                    executors
                    if distance == "l2"
                    else [distance_mapping[distance], *executors]
                )
            )
            if not group_match
            else GroupExperiment(
                executors=(
                    executors
                    if distance == "l2"
                    else [distance_mapping[distance], *executors]
                ),
                reporter=MatchingDatasetReporter(),
            )
        )

    def __init__(
        self,
        group_match: bool = False,
        distance: Literal["mahalanobis", "l2"] = "mahalanobis",
        metric: Literal["atc", "att", "ate"] = "ate",
        bias_estimation: bool = True,
        quality_tests: (
            Literal["smd", "psi", "ks-test", "repeats", "t-test", "auto"]
            | list[Literal["smd", "psi", "ks-test", "repeats", "t-test", "auto"]]
        ) = "auto",
        faiss_mode: Literal["base", "fast", "auto"] = "auto",

    ):
        super().__init__(
            experiment=self._make_experiment(
                group_match, distance, metric, bias_estimation, quality_tests, faiss_mode
            ),
            output=MatchingOutput(GroupExperiment if group_match else MatchingAnalyzer),
        )



from __future__ import annotations

from typing import Any

from ..analyzers.matching import MatchingAnalyzer
from ..dataset import (
    AdditionalMatchingRole,
    Dataset,
    ExperimentData,
    GroupingRole,
    StatisticRole,
    TargetRole,
)
from ..reporters.matching import MatchingDictReporter, MatchingQualityDatasetReporter
from ..utils import ID_SPLIT_SYMBOL, MATCHING_INDEXES_SPLITTER_SYMBOL
from .base import Output


class MatchingOutput(Output):
    resume: Dataset
    full_data: Dataset
    quality_results: Dataset

    def __init__(self, searching_class: type = MatchingAnalyzer):
        super().__init__(
            resume_reporter=MatchingDictReporter(searching_class),
            additional_reporters=MatchingQualityDatasetReporter(),
        )

    def _extract_full_data(self, experiment_data: ExperimentData, indexes: Dataset):
        indexes.index = experiment_data.ds.index
        filtered_field = indexes.drop(
            indexes[indexes[indexes.columns[0]] == -1], axis=0
        )
        matched_data = experiment_data.ds.loc[
            list(map(lambda x: x[0], filtered_field.get_values()))
        ].rename({i: i + "_matched" for i in experiment_data.ds.columns})
        matched_data.index = filtered_field.index
        self.indexes = indexes
        self.full_data = experiment_data.ds.append(
            matched_data.reindex(experiment_data.ds.index), axis=1
        )

    def extract(self, experiment_data: ExperimentData):
        resume = self.resume_reporter.report(experiment_data)
        reformatted_resume: dict[str, Any] = {}
        for key, value in resume.items():
            if ID_SPLIT_SYMBOL in key:
                keys = key.split(ID_SPLIT_SYMBOL)
                temp_key = keys[0] if len(keys) < 3 else f"{keys[2]} {keys[0]}"
                if temp_key not in reformatted_resume:
                    reformatted_resume[temp_key] = {}
                reformatted_resume[temp_key].update({keys[1]: value})
        if "indexes" in reformatted_resume.keys():
            group_indexes_id = experiment_data.ds.search_columns(GroupingRole())
            indexes = [
                Dataset.from_dict(
                    {
                        "indexes": list(
                            map(int, values.split(MATCHING_INDEXES_SPLITTER_SYMBOL))
                        )
                    },
                    index=experiment_data.ds[
                        experiment_data.ds[group_indexes_id] == group
                    ].index,
                    roles={"indexes": StatisticRole()},
                )
                for group, values in reformatted_resume.pop("indexes").items()
            ]
            indexes = indexes[0].append(indexes[1:]).sort()
        else:
            indexes = Dataset.from_dict(
                {
                    "indexes": list(
                        map(
                            int,
                            resume["indexes"].split(MATCHING_INDEXES_SPLITTER_SYMBOL),
                        )
                    )
                },
                roles={"indexes": AdditionalMatchingRole()},
            )

        outcome = experiment_data.field_search(TargetRole())[0]
        reformatted_resume["outcome"] = {
            key: outcome
            for key in reformatted_resume[next(iter(reformatted_resume.keys()))].keys()
        }

        self.resume = Dataset.from_dict(
            reformatted_resume,
            roles={
                column: StatisticRole() for column in list(reformatted_resume.keys())
            },
        )
        self._extract_full_data(
            experiment_data,
            indexes,
        )
        self.resume = round(self.resume, 2)

        self.quality_results = self.additional_reporters.report(experiment_data)



from __future__ import annotations

from typing import Any, Literal, Sequence

from ..dataset.dataset import Dataset, ExperimentData
from ..dataset.roles import (
    FeatureRole,
)
from ..utils import ScalarType
from ..utils.adapter import Adapter
from .abstract import Transformer


class NaFiller(Transformer):
    def __init__(
        self,
        target_roles: str | Sequence[str] | None = None,
        values: ScalarType | dict[str, ScalarType] | None = None,
        method: Literal["bfill", "ffill"] | None = None,
        key: Any = "",
    ):
        """
        Initializes a NaFiller object.

        Args:
            target_roles (Optional[Union[str, Sequence[str]]], optional): The roles of the target columns. Defaults to None.
            key (Any, optional): The key for the NaFiller object. Defaults to "".
            values (Union[ScalarType, Dict[str, ScalarType]], optional): The values to fill missing values with. Defaults to None.
            method (Literal["bfill", "ffill"], optional): The method to fill missing values. Defaults to None.

        Returns:
            None
        """

        super().__init__(key=key)
        self.target_roles = target_roles or FeatureRole()
        self.values = values
        self.method = method

    @staticmethod
    def _inner_function(
        data: Dataset,
        target_cols: str | None = None,
        values: ScalarType | dict[str, ScalarType] | None = None,
        method: Literal["bfill", "ffill"] | None = None,
    ) -> Dataset:
        target_cols = Adapter.to_list(target_cols)
        for column in target_cols:
            value = values[column] if isinstance(values, dict) else values
            data[column] = data[column].fillna(values=value, method=method)
        return data

    def execute(self, data: ExperimentData) -> ExperimentData:
        target_cols = data.ds.search_columns(roles=self.target_roles)
        result = data.copy(
            data=self.calc(
                data=data.ds,
                target_cols=target_cols,
                values=self.values,
                method=self.method,
            )
        )
        return result



from __future__ import annotations

from typing import Any, ClassVar

from ..analyzers.matching import MatchingAnalyzer
from ..comparators import KSTest, TTest
from ..dataset import (
    Dataset,
    ExperimentData,
)
from ..ml import FaissNearestNeighbors
from ..reporters.abstract import DatasetReporter, DictReporter, TestDictReporter
from ..utils import (
    ID_SPLIT_SYMBOL,
    MATCHING_INDEXES_SPLITTER_SYMBOL,
    ExperimentDataEnum,
)


class MatchingDictReporter(DictReporter):
    def __init__(self, searching_class: type = MatchingAnalyzer):
        self.searching_class = searching_class
        super().__init__()

    @staticmethod
    def _convert_dataset_to_dict(data: Dataset) -> dict[str, Any]:
        dict_data = data.to_dict()["data"]
        indexes = dict_data["index"]
        df = dict_data["data"]
        result = {}
        for key, values in df.items():
            for index, value in zip(indexes, values):
                result[f"{key}{ID_SPLIT_SYMBOL}{index}"] = value
        return result

    def _extract_from_analyser(self, data: ExperimentData):
        analyzer_id = data.get_one_id(
            self.searching_class, ExperimentDataEnum.analysis_tables
        )
        return self._convert_dataset_to_dict(data.analysis_tables[analyzer_id])

    @staticmethod
    def _extract_from_additional_fields(data: ExperimentData):
        indexes_id = data.get_one_id(
            FaissNearestNeighbors, ExperimentDataEnum.additional_fields
        )
        return {
            "indexes": MATCHING_INDEXES_SPLITTER_SYMBOL.join(
                str(i)
                for i in data.additional_fields[indexes_id].to_dict()["data"]["data"][
                    indexes_id
                ]
            )
        }

    def report(self, experiment_data: ExperimentData):
        result = {}
        result.update(self._extract_from_analyser(experiment_data))
        if self.searching_class == MatchingAnalyzer:
            result.update(self._extract_from_additional_fields(experiment_data))
        return result


class MatchingQualityDictReporter(TestDictReporter):
    tests: ClassVar[list] = [TTest, KSTest]

    def report(self, data: ExperimentData) -> dict[str, Any]:
        return self.extract_tests(data)


class MatchingQualityDatasetReporter(MatchingQualityDictReporter):

    @classmethod
    def convert_flat_dataset(cls, data: dict) -> Dataset:
        struct_dict = cls._get_struct_dict(data)
        return cls._convert_struct_dict_to_dataset(struct_dict)

    def report(self, data: ExperimentData):
        front_buffer = self.front
        self.front = False
        dict_report = super().report(data)
        self.front = front_buffer
        result = self.convert_flat_dataset(dict_report)
        return result


class MatchingDatasetReporter(DatasetReporter):
    def __init__(self, searching_class: type = MatchingAnalyzer) -> None:
        self.dict_reporter = MatchingDictReporter(searching_class)
        super().__init__(self.dict_reporter)



from __future__ import annotations

from typing import Any, ClassVar

from ..analyzers.ab import ABAnalyzer
from ..comparators import Chi2Test, TTest, UTest
from ..dataset import Dataset, ExperimentData
from ..utils import ExperimentDataEnum
from .aa import OneAADictReporter


class ABDictReporter(OneAADictReporter):
    tests: ClassVar[list] = [TTest, UTest, Chi2Test]

    def extract_analyzer_data(self, data: ExperimentData) -> dict[str, Any]:
        analyzer_id = data.get_one_id(ABAnalyzer, ExperimentDataEnum.analysis_tables)
        return self.extract_from_one_row_dataset(data.analysis_tables[analyzer_id])

    def extract_data_from_analysis_tables(self, data: ExperimentData) -> dict[str, Any]:
        result = {}
        result.update(self.extract_group_sizes(data))
        result.update(self.extract_group_difference(data))
        result.update(self.extract_tests(data))
        result.update(self.extract_analyzer_data(data))
        return result

    def report(self, data: ExperimentData) -> dict[str, Any]:
        return self.extract_data_from_analysis_tables(data)


class ABDatasetReporter(ABDictReporter):
    @staticmethod
    def _invert_aa_format(table: Dataset) -> Dataset:
        return table.replace("NOT OK", "N").replace("OK", "NOT OK").replace("N", "OK")

    def report(self, data: ExperimentData):
        front_buffer = self.front
        self.front = False
        dict_report = super().report(data)
        self.front = front_buffer
        result = self.convert_flat_dataset(dict_report)
        return self._invert_aa_format(result)



from __future__ import annotations

from typing import Any, Sequence

from ..dataset.dataset import Dataset, ExperimentData
from ..dataset.roles import (
    FeatureRole,
)
from ..utils import CategoricalTypes
from ..utils.adapter import Adapter
from .abstract import Transformer


class CategoryAggregator(Transformer):
    def __init__(
        self,
        target_roles: str | Sequence[str] | None = None,
        threshold: int | None = 15,
        new_group_name: str | None = None,
        key: Any = "",
    ):
        super().__init__(key=key)
        self.target_roles = target_roles or FeatureRole()
        self.threshold = threshold
        self.new_group_name = new_group_name

    @property
    def search_types(self):
        return [CategoricalTypes]

    @staticmethod
    def _inner_function(
        data: Dataset,
        target_cols: str | None = None,
        threshold: int | None = 15,
        new_group_name: str | None = None,
    ) -> Dataset:
        target_cols = Adapter.to_list(target_cols)
        for column in target_cols:
            categories_counts = data[column].value_counts()
            values_to_replace = categories_counts[
                categories_counts["count"] < threshold
            ][column].get_values(column=column)
            data[column] = data[column].replace(
                to_replace=values_to_replace, value=new_group_name
            )

        return data

    def execute(self, data: ExperimentData) -> ExperimentData:
        target_cols = data.ds.search_columns(
            roles=self.target_roles, search_types=self.search_types
        )
        result = data.copy(
            data=self.calc(
                data=data.ds,
                target_cols=target_cols,
                threshold=self.threshold,
                new_group_name=self.new_group_name,
            )
        )
        return result
